# ------------
# QPACE / qrun
# ------------
#
# Global constants and variables
#
# $Id: partition.pm 3569 2010-05-08 17:11:08Z pleiter $
#---------------------------------------------------------------------------------------------------

package Partition;

use strict;
use XML::XQL;
use XML::XQL::DOM;

use node;
use logger;

use defs qw(BLUE GREEN RED);
use defs qw(%bcMap);

#---------------------------------------------------------------------------------------------------
# Constructor
#---------------------------------------------------------------------------------------------------

sub new($$$)
{
  my $class = shift;
  my ($nc0Str, $nc1Str, $bc)  = @_;

  my $self = {};
  $self->{name}  = undef;
  $self->{ncLst} = [];
  $self->{size}  = [];

  # Initialize nc0: absolute position of node-card at partition origin
  if ($nc0Str =~ /^\s*(\d+):(\d+)\s*$/) {
    $self->{nc0} = Node->new($1, $2);
  }
  else {
    die "Illegal format for origin: \"$nc0Str\"\n";
  }

  # Initialize nc1: opposite corner
  if ($nc1Str =~ /^\s*(\d+):(\d+)\s*$/) {
    $self->{nc1} = Node->new($1, $2);
  }
  else {
    die "Illegal format for origin: \"$nc1Str\"\n";
  }

  if (defined $bcMap{$bc}) {
    $self->{bc} = $bcMap{$bc};
  }
  elsif (grep($bcMap{$_} == $bc, keys %bcMap)) {
    $self->{bc} = $bc;
  }
  else {
    die "Illegal argument for option bc: $bc\n";
  }

  bless ($self, $class);

  # Lookup partition name
  $self->_getPartitionName();

  # Create list of nodes
  $self->_createNcLst();

  logger::debug("size  = (%d,%d,%d)\n", @{$self->{size}});
  foreach my $node (@{$self->{ncLst}}) {
    logger::debug("%s -> %s\n", $node->hostname, $node->cpos_str);
  }

  return $self;
}


#---------------------------------------------------------------------------------------------------
# Initialize topology information (similar to libtnw/tnw_ppu.c)
#---------------------------------------------------------------------------------------------------

sub _createNcLst
{
  my $self = shift;
  my @size;

  #-------------------------------------------------------------------------------------------------
  # Initialise size: red direction
  #-------------------------------------------------------------------------------------------------

  $size[RED] = ($self->{nc1}->{nc}%8) - ($self->{nc0}->{nc}%8) + 1;

  # Sanity checks
  if ($size[RED] == 8) {
    if (($self->{nc0}->{nc} % 8) != 0) {
      die "Illegal partition with r_size=8\n";
    }
  }
  elsif ($size[RED] == 4) {
    if (($self->{nc0}->{nc} % 4) != 0) {
      die "Illegal partition with r_size=4\n";
    }
  }
  elsif ($size[RED] != 2 && $size[RED] != 1) {
    die "Illegal partition with r_size=$size[RED]\n";
  }

  #-------------------------------------------------------------------------------------------------
  # Initialise size: green direction
  #-------------------------------------------------------------------------------------------------

  my $bplr0 = (($self->{nc0}->{nc} % 16) < 8) ? 0 : 1;	# node slot in left/right backplane half
  my $bplr1 = (($self->{nc1}->{nc} % 16) < 8) ? 0 : 1;

  my $bptb0 = ($self->{nc0}->{nc} < 16) ? 0 : 1;	# node slot in top/bottom backplane half
  my $bptb1 = ($self->{nc1}->{nc} < 16) ? 0 : 1;

  my $dhg01 = ($bplr0 == $bplr1) ? 0 : 1;
  my $dvg01 = (2 * (($self->{nc1}->{bp} % 4) - ($self->{nc0}->{bp} % 4) + 1) - $bptb0 - (1- $bptb1)) - 1;

  $size[GREEN] = ($dhg01 + 1) * ($dvg01 + 1);

  # Sanity checks
  if ($size[GREEN] != 1 && $size[GREEN] != 2 && ($size[GREEN] % 4) != 0) {
    die "Illegal partition with g_size=$size[GREEN]\n"
  }

  #-------------------------------------------------------------------------------------------------
  # Initialise size: blue direction
  #-------------------------------------------------------------------------------------------------

  my $ra0 = int($self->{nc0}->{bp} / 8);
  my $ra1 = int($self->{nc1}->{bp} / 8);

  my $rafb0 = ($self->{nc0}->{bp} % 8) < 4 ? 0 : 1;
  my $rafb1 = ($self->{nc1}->{bp} % 8) < 4 ? 0 : 1;

  my $dhb01 = ($rafb0 == $rafb1) ? 0 : 1;
  my $dvb01 = $ra1 - $ra0;

  $size[BLUE] = ($dhb01 + 1) * ($dvb01 + 1);

  # Sanity checks
  if ($size[BLUE] != 1 && ($size[BLUE] % 2) != 0) {
    die "Illegal partition with b_size=$size[BLUE]\n"
  }

  #-------------------------------------------------------------------------------------------------
  # Update colour coordinates of nc0 and nc1
  #-------------------------------------------------------------------------------------------------

  $self->{nc0}->setCpos(0,0,0);
  $self->{nc1}->setCpos($size[BLUE]-1, $size[GREEN]-1, $size[RED]-1);

  logger::debug("nc0   = %s -> %s\n", $self->{nc0}->hostname, $self->{nc0}->cpos_str);
  logger::debug("nc1   = %s -> %s\n", $self->{nc1}->hostname, $self->{nc1}->cpos_str);
  logger::debug("size  = (%d,%d,%d)\n", @size);

  #-------------------------------------------------------------------------------------------------
  # Generate node-list
  #-------------------------------------------------------------------------------------------------

  foreach my $bp ($self->{nc0}->{bp} .. $self->{nc1}->{bp}) {
    foreach my $nc ($self->{nc0}->{nc} .. $self->{nc1}->{nc}) {
      my @x = (undef, undef, undef);			# node position (colour coordinates)

      #logger::debug("try nc-%02d-%02d\n", $bp, $nc);

      #---------------------------------------------------------------------------------------------
      # Red direction
      #---------------------------------------------------------------------------------------------

      $x[RED] = ($nc%8) - ($self->{nc0}->{nc}%8);

      next unless (0 <= $x[RED] && $x[RED] < $size[RED]);

      #---------------------------------------------------------------------------------------------
      # Green direction
      #---------------------------------------------------------------------------------------------

      my $bplr2 = (($nc  % 16) < 8) ? 0 : 1;
      my $bptb2 = ($nc  < 16) ? 0 : 1;

      my $dhg = ($bplr0 == $bplr2) ? 0 : 1;
      my $dvg = (2 * (($bp % 4) - ($self->{nc0}->{bp} % 4) + 1) - $bptb0 - (1- $bptb2)) - 1;

      $x[GREEN] = ($dvg == 0 || $dhg == 1) ? $dvg + $dhg : $size[GREEN] - ($dvg + $dhg);

      next unless (0 <= $dhg && $dhg <= $dhg01);
      next unless (0 <= $dvg && $dvg <= $dvg01);
      next unless (0 <= $x[GREEN] && $x[GREEN] < $size[GREEN]);

      #---------------------------------------------------------------------------------------------
      # Blue direction
      #---------------------------------------------------------------------------------------------

      my $ra2   = int($bp / 8);
      my $rafb2 = ($bp % 8) < 4 ? 0 : 1;

      my $dhb = ($rafb0 == $rafb2) ? 0 : 1;
      my $dvb = $ra2 - $ra0;

      $x[BLUE] = ($dvb == 0 || $dhb == 1) ? $dvb + $dhb : $size[BLUE] - ($dvb + $dhb);

      next unless (0 <= $dhb && $dhb <= $dhb01);
      next unless (0 <= $dvb && $dvb <= $dvb01);
      next unless (0 <= $x[BLUE] && $x[BLUE] < $size[BLUE]);

      my $node = Node->new($bp, $nc, @x);
      push @{$self->{ncLst}}, $node;
      #logger::debug("%s -> %s\n", $node->hostname, $node->cpos_str);
    }
  }

  # Sanity check
  if ($size[BLUE]*$size[GREEN]*$size[RED] != scalar @{$self->{ncLst}}) {
    die "Partition size and node list length mismatch\n";
  }

  push @{$self->{size}}, @size;
}


#---------------------------------------------------------------------------------------------------
# Find partition name
#---------------------------------------------------------------------------------------------------

sub _getPartitionName
{
  my $self = shift;

  # Find file with partition names
  return unless defined $ENV{QROOT};

  chomp(my $hostname = `hostname`);
  my $partFile = undef;
  for ($hostname) {
    /qmasterJ/ && do { $partFile = "$ENV{QROOT}/etc/ju-part.xml"; last; };
    /qmasterW/ && do { $partFile = "$ENV{QROOT}/etc/wu-part.xml"; last; };
  }

  return unless (defined $partFile && -f $partFile);

  # Parse file
  my $parser = new XML::DOM::Parser;
  my $doc = $parser->parsefile($partFile);
  die "Failed to parse \"$partFile\"\n" unless defined $doc;

  my $expr = "/qpartition/partition[upperleft=\""  . $self->{nc0}->hostname . "\" and " .
	                           "lowerright=\"" . $self->{nc1}->hostname . "\"]/name";
  logger::debug("Expr = $expr\n");
  my $query = new XML::XQL::Query(Expr => $expr);
  my @result = $query->solve($doc);
  $query->dispose;

  if (scalar @result == 1) {
    $self->{name} = $result[0]->getChildNodes->item(0)->getNodeValue;
    logger::debug("Partition name: \"%s\"\n", $self->{name});
  }
  elsif (scalar @result > 1) {
    die "Multiple entries in \"$partFile\" match (contact admins!)\n";
  }
  else {
    $self->{name} = undef;
    logger::info("Partition not defined in \"$partFile\"\n");
  }
}

#---------------------------------------------------------------------------------------------------
# Find vertices from name
#---------------------------------------------------------------------------------------------------

sub _getPartitionVertices
{
  my $self = shift;

  # Find file with partition names
  return unless defined $ENV{QROOT};

  chomp(my $hostname = `hostname`);
  my $partFile = undef;
  for ($hostname) {
    /qmasterJ/ && do { $partFile = "$ENV{QROOT}/etc/ju-part.xml"; last; };
    /qmasterW/ && do { $partFile = "$ENV{QROOT}/etc/wu-part.xml"; last; };
  }

  return unless (defined $partFile && -f $partFile);

  # Parse file
  my $parser = new XML::DOM::Parser;
  my $doc = $parser->parsefile($partFile);
  die "Failed to parse \"$partFile\"\n" unless defined $doc;

  my $expr = "/qpartition/partition[name='" . $self->{name} . "']";
  logger::debug("Expr = $expr\n");
  my $query = new XML::XQL::Query(Expr => $expr);
  my @result = $query->solve($doc);
  $query->dispose;

  if (scalar @result == 1) {
    my @cns = $result[0]->getChildNodes();
    for my $cn (@cns) {
      if ($cn->getNodeType() == XML::DOM::ELEMENT_NODE) {
        if ($cn->getTagName() eq "upperleft") {
          my $ncs = $cn->getChildNodes()->item(0)->getNodeValue();
          if ($ncs =~ /^nc-(\d\d)-(\d\d)$/) {
            $self->{nc0} = Node->new($1, $2);
          }
          else {
            die "Illegal format for origin: \"$ncs\"\n";
          }
        }
        if ($cn->getTagName() eq "lowerright") {
          my $ncs = $cn->getChildNodes()->item(0)->getNodeValue() . "\n";
          if ($ncs =~ /^nc-(\d\d)-(\d\d)$/) {
            $self->{nc1} = Node->new($1, $2);
          }
          else {
            die "Illegal format for origin: \"$ncs\"\n";
          }
        }
      }
    }
  }
  elsif (scalar @result > 1) {
    die "Multiple entries in \"$partFile\" match (contact admins!)\n";
  }
  else {
    die "Partition not defined in \"$partFile\"\n";
  }

  if (scalar @result == 1) {
    $self->{name} = $result[0]->getChildNodes->item(0)->getNodeValue;
    logger::debug("Partition name: \"%s\"\n", $self->{name});
  }
  elsif (scalar @result > 1) {
    die "Multiple entries in \"$partFile\" match (contact admins!)\n";
  }
  else {
    $self->{name} = undef;
    logger::info("Partition not defined in \"$partFile\"\n");
  }
}

#---------------------------------------------------------------------------------------------------
# Find boundary conditions of partition
#---------------------------------------------------------------------------------------------------

sub _getPartitionBC
{
  my $self = shift;

  return unless defined $ENV{QROOT};

  chomp(my $hostname = `hostname`);
  my $cablesFile = undef;
  for ($hostname) {
    /qmasterJ/ && do { $cablesFile = "$ENV{QROOT}/etc/ju-cables.xml"; last; };
    /qmasterW/ && do { $cablesFile = "$ENV{QROOT}/etc/wu-cables.xml"; last; };
  }

  return unless (defined $cablesFile && -f $cablesFile);

  my $parser = new XML::DOM::Parser;
  my $doc = $parser->parsefile($cablesFile);
  die "Failed to parse \"$cablesFile\"\n" unless defined $doc;

  #get rack number and level from upper left node
  $self->{nc0}->hostname =~ /nc-(..)-../;
  my $backplane = $1;
  my $rackno = int($backplane / 8);
  my $level  = $backplane % 4;

  my $expr = "/qbluecab/boundary[\@rack =~ '/[WJwj]{0,1}$rackno/' \$and\$ \@level = '$level']";
  logger::debug("Expr = $expr\n");
  my $query = new XML::XQL::Query(Expr => $expr);
  my @result = $query->solve($doc);
  $query->dispose;

  if (scalar @result == 1) {
    logger::debug("found bc for partition\n");
  }
  elsif (scalar @result > 1) {
    die "Multiple entries in \"$cablesFile\" match.\n";
  }
  else {
    die "boundary conditions undefined in \"$cablesFile\"\n";
  }

  my $bctype = $result[0]->getAttribute('type');
  logger::debug("bc is: $bctype\n");

  if (defined $bcMap{$bctype}) {
    $self->{bc} = $bcMap{$bctype};
  }
  else {
    die "File $cablesFile uses invalid syntax";
  }
}

#---------------------------------------------------------------------------------------------------
# Initialize from vertices only.
#---------------------------------------------------------------------------------------------------

sub newFromVertices($$)
{
  my $class = shift;
  my ($nc0Str, $nc1Str)  = @_;

  my $self = {};
  $self->{name}  = undef;
  $self->{ncLst} = [];
  $self->{size}  = [];

  # Initialize nc0: absolute position of node-card at partition origin
  if ($nc0Str =~ /^\s*(\d+):(\d+)\s*$/) {
    $self->{nc0} = Node->new($1, $2);
  }
  else {
    die "Illegal format for origin: \"$nc0Str\"\n";
  }

  # Initialize nc1: opposite corner
  if ($nc1Str =~ /^\s*(\d+):(\d+)\s*$/) {
    $self->{nc1} = Node->new($1, $2);
  }
  else {
    die "Illegal format for origin: \"$nc1Str\"\n";
  }

  bless ($self, $class);

  # Lookup partition name
  $self->_getPartitionName();

  # Lookup partition boundary conditions
  $self->_getPartitionBC();

  # Create list of nodes
  $self->_createNcLst();

  logger::debug("size  = (%d,%d,%d)\n", @{$self->{size}});
  foreach my $node (@{$self->{ncLst}}) {
    logger::debug("%s -> %s\n", $node->hostname, $node->cpos_str);
  }

  return $self;
}

#---------------------------------------------------------------------------------------------------
# Initialize from name only.
#---------------------------------------------------------------------------------------------------

sub newFromName($)
{
  my $class = shift;
  my ($pname)  = @_;

  my $self = {};
  $self->{name}  = $pname;
  $self->{ncLst} = [];
  $self->{size}  = [];

  bless ($self, $class);

  # Lookup partitiion vertices
  $self->_getPartitionVertices();

  # Lookup partition boundary conditions
  $self->_getPartitionBC();

  # Create list of nodes
  $self->_createNcLst();

  logger::debug("size  = (%d,%d,%d)\n", @{$self->{size}});
  foreach my $node (@{$self->{ncLst}}) {
    logger::debug("%s -> %s\n", $node->hostname, $node->cpos_str);
  }

  return $self;
}


1;
