# ------------
# QPACE / qrun
# ------------
#
# Functions to execute remote operations
#
# $Id: remote.pm 3531 2010-04-21 15:22:04Z pleiter $
#---------------------------------------------------------------------------------------------------

package remote;

use strict;
use FileHandle;

use logger;
use defs qw($opt_dry);

my $RSH = "ssh -o ConnectTimeout=3";
my $RCP = "scp -o ConnectTimeout=3";


#---------------------------------------------------------------------------------------------------
# Remote execution
# - First argument is either reference to Node or reference to Partition
#---------------------------------------------------------------------------------------------------

sub exec($$)
{
  my ($addr, $rCmd) = @_;

  die "Illegal argument" unless (ref($addr) eq "REF");

  my $addRank = 0;
  my $nodeLstRef = undef;
  if (ref($$addr) eq "Node") {
    $nodeLstRef = [];
    push @{$nodeLstRef}, $$addr;
  }
  elsif (ref($$addr) eq "Partition") {
    $addRank = 1;
    $nodeLstRef = \@{$$addr->{ncLst}};
  }

  die "Illegal argument" unless defined $nodeLstRef;

  my @cpidLst = ();
  foreach my $i (0..$#{$nodeLstRef}) {
    my $node = $nodeLstRef->[$i];
    my $cmd = "$RSH root\@" . $node->hostname .
              " \'bash -c \"" . ($addRank ? "declare -x QRUN_RANK=$i; " : "") . "$rCmd\"\'";
    logger::debug("$cmd\n");

    if (!$node->isUp) {
      logger::debug("Node \"%s\" down [skipped]\n", $node->hostname);
    }
    elsif (!$opt_dry) {
      my $cpid = fork();
      if ($cpid > 0) {			# parent
        push @cpidLst, $cpid;
      }
      elsif ($cpid == 0) {		# child
        system("$cmd");
        exit(0);
      }
    }
  }

  foreach my $cpid (@cpidLst) {
    waitpid($cpid, 0);
  }
}


#---------------------------------------------------------------------------------------------------
# Copy to node
# - First argument is either reference to Node or reference to Partition
# - May include directories
#---------------------------------------------------------------------------------------------------

sub copy_to
{
  _copy("TO", @_);
}


#---------------------------------------------------------------------------------------------------
# Copy from node
# - First argument is either reference to Node or reference to Partition
# - May include directories
#---------------------------------------------------------------------------------------------------

sub copy_from
{
  _copy("FROM", @_);
}


#---------------------------------------------------------------------------------------------------
# Remote copy
# - First argument is either reference to Node or reference to Partition
# - May include directories
#---------------------------------------------------------------------------------------------------

sub _copy
{
  my $direction   = shift @_;
  my $addr        = shift @_;
  my $destination = pop @_;
  my @files       = @_;

  die "Illegal argument" unless (ref($addr) eq "REF");

  my $nodeLstRef = undef;
  if (ref($$addr) eq "Node") {
    $nodeLstRef = [];
    push @{$nodeLstRef}, $$addr;
  }
  elsif (ref($$addr) eq "Partition") {
    $nodeLstRef = \@{$$addr->{ncLst}};
  }

  die "Illegal argument" unless defined $nodeLstRef;

  my @cpidLst = ();
  foreach my $node (@{$nodeLstRef}) {
    my $cmd;
    if ($direction eq "TO") {
      $cmd = "$RCP -pqr " . join(" ", @files) . " root\@" . $node->hostname . ":$destination";
    }
    else {          # FROM
      $cmd = "$RCP -pqr root\@" . $node->hostname . ":$files[0] $destination";
    }
    logger::debug("$cmd\n");

    if (!$node->isUp) {
      logger::debug("Node \"%s\" down [skipped]\n", $node->hostname);
    }
    elsif (!$opt_dry) {
      my $cpid = fork();
      if ($cpid > 0) {			# parent
        push @cpidLst, $cpid;
      }
      elsif ($cpid == 0) {          # child
        system("$cmd");
        exit(0);
      }
    }
  }

  foreach my $cpid (@cpidLst) {
    waitpid($cpid, 0);
  }
}


#---------------------------------------------------------------------------------------------------
# Open pipe with remote execution command for reading
# - First argument must be reference to Node
#---------------------------------------------------------------------------------------------------

sub read_pipe
{
  my ($node, $cmd, $fh) = @_;

  $cmd = "$RSH root\@" . $node->hostname . " $cmd";
  logger::debug("$cmd\n");

  $$fh = undef;
  if (!$node->isUp) {
    logger::debug("Node \"%s\" down [skipped]\n", $node->hostname);
  }
  elsif (!$opt_dry) {
    $$fh = new FileHandle "$cmd |";
    die "Failed to open pipe for reading ($!)\n" unless defined $$fh;
  }
}


1;
